/*
 *  client/main.c  --  LysKOM client, main program
 *
 *
 *  Copyright (C) 1990  Lysator Computer Club
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  E-mail:	Bellman@Lysator.LiU.SE
 */


#include <stdio.h>
#include <stdlib.h>

#include <zmalloc.h>
#include <s-string.h>

#include <kom-types.h>
#include <services.h>

#include "version.h"

#include "error.h"
#include "read-line.h"
#include "defaults.h"
#include "texts.h"
#include "commands.h"
#include "functions.h"
#include "cache.h"
#include "xoutput.h"



Info		  kom_info;	/*  LysKOM server parameters  */

Old_tty		  save_tty_param;


int
main (int	   argc,
      char	** argv)

{
    static char		* this_function	= "main()";


    s_set_storage_management (zmalloc, zrealloc, zfree);
    kom_set_storage_management (zmalloc, zrealloc, zfree);
    init_messages ();	/* Initialize the 'txts' struct */
    init_cmd_list ();	/* Initialize the list of commands */
    init_terminal (stdin, stdout, &save_tty_param);
    init_cache ();

    if (kom_open_connection(argv[1] ? argv[1] : DEFAULT_SERVER) == FAILURE)
    {
	fatal_client_error(CLIENT_SERVER_ERROR,
			   this_function, "Can't connect to server",
			   (fatal_info) kom_errno);
	exit(1);
    }

    /*  OK, connection established  */
    kom_get_info (&kom_info);	/* BUG! Retval not checked */

    /*  Greet user  */
    xprintf("%s %s, %s v" VERSION ", %s v%5.2f\n",
	   txts.Welcome_to,
	   txts.The_Name,
	   txts.TTY_client,
	   txts.server,
	   (float)kom_info.version/10000);

    /*  Log in to server  */
    cmd_start_anew (EMPTY_STRING);
    
    {
	String 		message = EMPTY_STRING;

	s_strcat(&message, s_fcrea_str("this is a long string constant because I cant find an allocate."));

	sprintf(message.string, txts.i_am_doing_startup, VERSION);
	s_strlen(message) = strlen(message.string);

	kom_change_what_i_am_doing(message);

	s_clear(&message);
    }

    /*  And run the command loop  */
    command_loop ();

    /* Restore the terminal to the old settings */
    restore_terminal (stdin, stdout, &save_tty_param);

    return 0;	/*  LysKOM can *never* be wrong! */
}

